import numpy as np

try:
    from omegaconf import OmegaConf
except ModuleNotFoundError:
    OmegaConf = None

_DIST_KEYS = ("uniform", "normal", "choice", "loguniform")

def _is_dist_node(x) -> bool:
    """True if x is a dict like {'uniform': [a,b]} etc."""
    return isinstance(x, dict) and len(x) == 1 and next(iter(x.keys())) in _DIST_KEYS

def _sample_dist(spec: dict, rng: np.random.Generator):
    k = next(iter(spec.keys()))
    v = spec[k]

    match k:
        case "uniform":
            lo, hi = v
            return float(rng.uniform(lo, hi))

        case "normal":
            mu, sigma = v
            return float(rng.normal(mu, sigma))

        case "choice":
            # v is a list of options
            return v[int(rng.integers(0, len(v)))]

        case "loguniform":
            lo, hi = v
            # require positive bounds
            if lo <= 0 or hi <= 0:
                raise ValueError(f"loguniform bounds must be > 0, got {v}")
            return float(np.exp(rng.uniform(np.log(lo), np.log(hi))))

    raise ValueError(f"Unknown distribution: {k}")

def sample_config(cfg, rng: np.random.Generator, ignore_paths=()):
    """
    Recursively sample a config.

    - cfg can be an OmegaConf object or plain python container.
    - ignore_paths is a tuple/list of dotpaths to leave untouched (copied as-is).
      Example: ('time.frame_durations_s',)
    """
    # Convert OmegaConf -> plain containers once at the top
    if (
        OmegaConf is not None
        and not isinstance(cfg, (dict, list))
        and OmegaConf.is_config(cfg)
    ):
        cfg = OmegaConf.to_container(cfg, resolve=True)

    ignore_paths = tuple(ignore_paths)

    def rec(node, path: str):
        # If this exact path is ignored, return node as-is
        if path in ignore_paths:
            return node

        # Dict
        if isinstance(node, dict):
            # Distribution leaf
            if _is_dist_node(node):
                return _sample_dist(node, rng)

            # Recurse
            out = {}
            for kk, vv in node.items():
                child_path = f"{path}.{kk}" if path else kk
                out[kk] = rec(vv, child_path)
            return out

        # List
        if isinstance(node, list):
            # Keep list shape; allow lists to contain fixed values or dist nodes
            out = []
            for i, vv in enumerate(node):
                child_path = f"{path}[{i}]"
                out.append(rec(vv, child_path))
            return out

        # Scalar leaf (fixed)
        return node

    return rec(cfg, "")



import warnings

import numpy as np
import torch

from .io_utils import (
    convert_time_to_minutes,
    infer_device,
    infer_dtype,
    normalize_tacs_shape,
    require_1d,
    to_tensor,
)
from .math_utils import build_interpolation_grid, interp1d_linear
from .solver import lls_3k, lls_3k_vB, lls_4k, lls_4k_vB
from .types import ArrayLike, DeviceLike, Scalar


def fit_compartment_model(
    tacs: ArrayLike,
    aif: ArrayLike,
    time: ArrayLike,
    *,
    model: str = "3k_vB",
    vB: Scalar | ArrayLike | None = None,
    aif_time: ArrayLike | None = None,
    interpolate: bool = True,
    frame_duration: float = 1.0 / 60.0,
    time_unit: str = "min",
    device: DeviceLike | None = None,
    dtype: torch.dtype | None = None,
    reg: float = 1e-6,
    eps: float = 1e-8,
    column_scale: bool = True,
    nnls: bool = False,
    nnls_max_iter: int = 200,
    nnls_tol: float = 1e-6,
) -> tuple[np.ndarray, np.ndarray, float]:
    """High-level interface for LLS compartment fitting.

    Accepts NumPy or torch inputs, runs fitting in torch, and always returns NumPy
    outputs as:
        (kinetic_params, fitted_tacs, mse)
    """

    resolved_device = infer_device(device, tacs, aif, time, aif_time, vB)
    resolved_dtype = infer_dtype(dtype, tacs, aif, time, aif_time)

    tacs_t = normalize_tacs_shape(
        to_tensor(tacs, device=resolved_device, dtype=resolved_dtype)
    )
    aif_t = to_tensor(aif, device=resolved_device, dtype=resolved_dtype)
    t_tac = to_tensor(time, device=resolved_device, dtype=resolved_dtype)
    t_aif = (
        t_tac
        if aif_time is None
        else to_tensor(aif_time, device=resolved_device, dtype=resolved_dtype)
    )

    if aif_t.ndim != 1:
        raise ValueError(f"aif must be 1D. Got shape {tuple(aif_t.shape)}.")
    if tacs_t.size(0) != t_tac.numel():
        raise ValueError("time and tacs must share the same number of samples.")
    if aif_t.numel() != t_aif.numel():
        raise ValueError("aif and aif_time must share the same number of samples.")

    require_1d("time", t_tac)
    require_1d("aif_time", t_aif)

    t_tac_min = convert_time_to_minutes(
        t_tac, time_unit=time_unit, warn_on_long_minutes=True
    )
    t_aif_min = convert_time_to_minutes(t_aif, time_unit=time_unit)

    if interpolate:
        t_model = build_interpolation_grid(t_tac_min, t_aif_min, frame_duration)
        tacs_model = interp1d_linear(t_tac_min, tacs_t, t_model, eps=eps)
        aif_model = interp1d_linear(t_aif_min, aif_t, t_model, eps=eps)
    else:
        if t_tac_min.shape != t_aif_min.shape or not torch.allclose(
            t_tac_min, t_aif_min, atol=eps, rtol=0.0
        ):
            raise ValueError("time and aif_time must match when interpolate=False.")
        t_model = t_tac_min
        tacs_model = tacs_t
        aif_model = aif_t

    model_key = model.lower().replace("-", "_")
    fixed_vb_models = {"3k", "lls_3k", "4k", "lls_4k"}
    fitted_vb_models = {"3k_vb", "lls_3k_vb", "4k_vb", "lls_4k_vb"}
    valid_models = fixed_vb_models | fitted_vb_models
    if model_key not in valid_models:
        raise ValueError(
            "Invalid model. Choose one of: "
            "'3k', '3k_vB', '4k', '4k_vB', 'lls_3k', 'lls_3k_vB', 'lls_4k', "
            "'lls_4k_vB'."
        )

    if model_key in fixed_vb_models:
        if vB is None:
            raise ValueError("vB must be provided for fixed-vB models ('3k' and '4k').")
        vb_t = to_tensor(vB, device=resolved_device, dtype=resolved_dtype)
    else:
        if vB is not None:
            warnings.warn(
                f"Ignoring vB for model='{model}'. This model estimates vB.",
                stacklevel=2,
            )
        vb_t = None

    kwargs = dict(
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
    )
    if model_key in {"3k", "lls_3k"}:
        kinetic_params, fitted_model, mse = lls_3k(
            tacs_model, aif_model, t_model, vB=vb_t, **kwargs
        )
    elif model_key in {"4k", "lls_4k"}:
        kinetic_params, fitted_model, mse = lls_4k(
            tacs_model, aif_model, t_model, vB=vb_t, **kwargs
        )
    elif model_key in {"3k_vb", "lls_3k_vb"}:
        kinetic_params, fitted_model, mse = lls_3k_vB(
            tacs_model, aif_model, t_model, **kwargs
        )
    else:
        kinetic_params, fitted_model, mse = lls_4k_vB(
            tacs_model, aif_model, t_model, **kwargs
        )

    if interpolate:
        fitted_tacs = interp1d_linear(t_model, fitted_model, t_tac_min, eps=eps)
    else:
        fitted_tacs = fitted_model

    kinetic_params_np = kinetic_params.detach().cpu().numpy()
    fitted_tacs_np = fitted_tacs.detach().cpu().numpy()
    mse_np = mse.detach().cpu().numpy()
    return kinetic_params_np, fitted_tacs_np, mse_np



import torch


def trapezoidal_integral(y: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
    """
    Compute the trapezoidal integral of y with respect to t along the first dimension.

    y: (T, N) tensor of values to integrate
    t: (T,) tensor of time points corresponding to the first dimension of y

    Returns
    -------
    integral: (T, N) tensor of integrals for each voxel
    """
    integral = torch.zeros_like(y)
    integral[1:, ...] = torch.cumulative_trapezoid(y, t, dim=0)
    return integral


def quadratic_formula(
    A: torch.Tensor | float | int, B: torch.Tensor, C: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor]:
    discriminant = B**2 - 4 * A * C
    sqrt_discriminant = torch.sqrt(discriminant)
    x1 = (-B + sqrt_discriminant) / (2 * A)
    x2 = (-B - sqrt_discriminant) / (2 * A)
    return x1, x2


def interp1d_linear(
    x: torch.Tensor, y: torch.Tensor, x_new: torch.Tensor, *, eps: float
) -> torch.Tensor:
    if y.ndim not in {1, 2}:
        raise ValueError(f"y must be 1D or 2D. Got shape {tuple(y.shape)}.")

    x_new_clamped = x_new.clamp(min=x[0].item(), max=x[-1].item())

    idx = torch.searchsorted(x, x_new_clamped, right=True)
    idx = idx.clamp(1, x.numel() - 1)

    x0 = x[idx - 1]
    x1 = x[idx]
    alpha = (x_new_clamped - x0) / (x1 - x0).clamp_min(eps)

    if y.ndim == 1:
        y0 = y[idx - 1]
        y1 = y[idx]
        return y0 + alpha * (y1 - y0)

    y0 = y[idx - 1, :]
    y1 = y[idx, :]
    return y0 + alpha[:, None] * (y1 - y0)


def build_interpolation_grid(
    t_tac: torch.Tensor, t_aif: torch.Tensor, frame_duration_min: float
) -> torch.Tensor:
    if frame_duration_min <= 0.0:
        raise ValueError("frame_duration must be > 0.")

    t_start = torch.maximum(t_tac[0], t_aif[0])
    t_end = torch.minimum(t_tac[-1], t_aif[-1])
    if t_end <= t_start:
        raise ValueError("No temporal overlap between TAC and AIF time ranges.")

    step = t_tac.new_tensor(frame_duration_min)
    n_steps = int(torch.round((t_end - t_start) / step).item())
    n_points = max(2, n_steps + 1)
    return torch.linspace(
        t_start.item(), t_end.item(), n_points, device=t_tac.device, dtype=t_tac.dtype
    )



import json
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.dataset as ds

T = 42  # number of frames


def write_time_sidecar(out_dir: str, seed: int):
    out = Path(out_dir)
    out.mkdir(parents=True, exist_ok=True)
    payload = {
        "seed": seed,
    }
    (out / "_metadata.json").write_text(json.dumps(payload, indent=2))


def make_table(
    sample_id: np.ndarray,
    factors: dict,  # e.g. {"protocol": ..., "noise": ...} arrays length N
    theta: dict,  # e.g. {"K1":..., "k2":...} arrays length N
    aif_params: dict,  # arrays length N
    aif: np.ndarray,  # (N, T) float32/float64
    tac: np.ndarray,  # (N, T)
    time: np.ndarray,  # (N, T)
):
    N = aif.shape[0]
    assert aif.shape == (N, T)
    assert tac.shape == (N, T)
    assert time.shape == (N, T)

    cols = {
        "sample_id": pa.array(sample_id),
    }

    # factors + params as scalar columns
    for k, v in factors.items():
        cols[k] = pa.array(v)
    for k, v in theta.items():
        cols[k] = pa.array(v)
    for k, v in aif_params.items():
        cols[k] = pa.array(v)

    # wide curve columns
    for j in range(T):
        cols[f"aif_{j:03d}"] = pa.array(aif[:, j])
        cols[f"tac_{j:03d}"] = pa.array(tac[:, j])
        cols[f"time_{j:03d}"] = pa.array(time[:, j])

    return pa.table(cols)


def write_parquet_dataset(
    table: pa.Table,
    root_dir: str,
    partition_cols=("protocol", "noise"),
    compression="zstd",
):
    root = Path(root_dir) / "parquet"
    root.mkdir(parents=True, exist_ok=True)

    ds.write_dataset(
        data=table,
        base_dir=str(root),
        format="parquet",
        partitioning=list(partition_cols),
        existing_data_behavior="overwrite_or_ignore",  # safe for repeated writes
        file_options=ds.ParquetFileFormat().make_write_options(
            compression=compression,
            use_dictionary=True,
            write_statistics=True,
        ),
        max_rows_per_file=50_000,  # tune
        max_rows_per_group=10_000,  # tune
    )


def batch_to_table(batch):
    # batch is a dict of numpy arrays length N (and curves are (N,T))
    N = batch["tac"].shape[0]
    cols = {
        "batch_id": pa.array(batch["batch_id"]),
        "sample_id": pa.array(batch["sample_id"]),
        "protocol": pa.array(batch["protocol"]),
        "noise": pa.array(batch["noise"]),
        "regime": pa.array(batch["regime"]),
        "n_frames_used": pa.array(batch["n_frames_used"]),
        # params...
        "K1": pa.array(batch["K1"]),
        "k2": pa.array(batch["k2"]),
        "k3": pa.array(batch["k3"]),
        "k4": pa.array(batch["k4"]),
        "vB": pa.array(batch["vB"]),
        # aif params...
        "Tinj_s": pa.array(batch["Tinj_s"]),
        "tau": pa.array(batch["tau"]),
        "A1": pa.array(batch["A1"]),
        "r2": pa.array(batch["r2"]),
        "r3": pa.array(batch["r3"]),
        "L1": pa.array(batch["L1"]),
        "L2": pa.array(batch["L2"]),
        "L3": pa.array(batch["L3"]),
    }

    aif = batch["aif"].astype(np.float32, copy=False)
    tac = batch["tac"].astype(np.float32, copy=False)
    time = batch["time"].astype(np.float32, copy=False)
    for j in range(T):
        cols[f"aif_{j:03d}"] = pa.array(aif[:, j])
        cols[f"tac_{j:03d}"] = pa.array(tac[:, j])
        cols[f"time_{j:03d}"] = pa.array(time[:, j])

    return pa.table(cols)


def write_batch(
    table: pa.Table,
    out_dir: str,
    partition_cols=("batch_id", "protocol", "noise", "regime"),
    compression="zstd",
):
    base_dir = str(Path(out_dir) / "parquet")
    ds.write_dataset(
        table,
        base_dir=base_dir,
        format="parquet",
        partitioning=list(partition_cols),
        partitioning_flavor="hive",
        existing_data_behavior="overwrite_or_ignore",
        file_options=ds.ParquetFileFormat().make_write_options(
            compression=compression,
            use_dictionary=True,
            write_statistics=True,
        ),
        max_rows_per_file=50_000,
        max_rows_per_group=10_000,
    )



from .high_level import fit_compartment_model
from .math_utils import trapezoidal_integral
from .models import make_compartment_model
from .solver import lls_3k, lls_3k_vB, lls_4k, lls_4k_vB

__all__ = [
    "lls_3k",
    "lls_3k_vB",
    "lls_4k",
    "lls_4k_vB",
    "fit_compartment_model",
    "make_compartment_model",
    "trapezoidal_integral",
]



import numpy as np


def feng_aif(t, tau, A1, A2, A3, L1, L2, L3):
    cp = (
        (A1 * (t - tau) - A2 - A3) * np.exp(-L1 * (t - tau))
        + A2 * np.exp(-L2 * (t - tau))
        + A3 * np.exp(-L3 * (t - tau))
    )
    cp[: np.sum(t < tau)] = 0
    return cp


def frame_average(y, t_s, frame_durations_s):
    edges = np.concatenate([[0], np.cumsum(frame_durations_s)])
    out = np.empty(len(frame_durations_s))
    for i, (a, b) in enumerate(zip(edges[:-1], edges[1:])):
        m = (t_s >= a) & (t_s < b)
        if np.any(m):
            out[i] = np.trapz(y[m], t_s[m]) / (b - a)
        else:
            out[i] = np.interp(0.5 * (a + b), t_s, y)
    return out


def generate_aif(
    t_fine_min,
    t_fine_s,
    frame_durations_s,
    dt_s,
    *,
    Tinj_s,
    tau,
    A1,
    r2,
    r3,
    L1,
    L2,
    L3,
):
    k = max(1, int(round(Tinj_s / dt_s)))
    kernel = np.ones(k) / k
    A2, A3 = r2 * A1, r3 * A1

    cp = feng_aif(t_fine_min, tau, A1, A2, A3, L1, L2, L3)
    cp_inj = np.convolve(cp, kernel, mode="same")
    return frame_average(cp_inj, t_fine_s, frame_durations_s)


def generate_time(*, frame_durations_s, dt_s):
    durs = []
    for n_frames, frame_dur in frame_durations_s:
        durs.append(np.full(n_frames, frame_dur))

    frame_durations_s = np.concatenate(durs)
    edges_s = np.concatenate([[0], np.cumsum(frame_durations_s)])
    mids_s = 0.5 * (edges_s[:-1] + edges_s[1:])

    t_fine_s = np.arange(0, edges_s[-1] + dt_s, dt_s)
    t_fine_min = t_fine_s / 60
    time_min = mids_s / 60

    return t_fine_min, t_fine_s, time_min, frame_durations_s, dt_s


def gaussian_noise(signal, rng, *, factor):
    amplitude_noise = rng.normal(loc=0, scale=signal)
    return signal + factor * amplitude_noise



import warnings

import torch

from .types import ArrayLike, DeviceLike, Scalar


def infer_device(
    device: DeviceLike | None, *values: ArrayLike | Scalar | None
) -> torch.device:
    if device is not None:
        return torch.device(device)
    for value in values:
        if torch.is_tensor(value):
            return value.device
    return torch.device("cpu")


def infer_dtype(dtype: torch.dtype | None, *values: ArrayLike | None) -> torch.dtype:
    if dtype is not None:
        return dtype
    for value in values:
        if torch.is_tensor(value) and value.dtype.is_floating_point:
            return value.dtype
    return torch.float32


def to_tensor(
    value: ArrayLike | Scalar, *, device: torch.device, dtype: torch.dtype
) -> torch.Tensor:
    if torch.is_tensor(value):
        return value.to(device=device, dtype=dtype)
    return torch.as_tensor(value, device=device, dtype=dtype)


def require_1d(name: str, value: torch.Tensor) -> None:
    if value.ndim != 1:
        raise ValueError(f"{name} must be 1D. Got shape {tuple(value.shape)}.")
    if value.numel() < 2:
        raise ValueError(f"{name} must have at least 2 time points.")
    if not torch.isfinite(value).all():
        raise ValueError(f"{name} contains non-finite values.")
    if torch.any(value[1:] <= value[:-1]):
        raise ValueError(f"{name} must be strictly increasing.")


def normalize_tacs_shape(tacs: torch.Tensor) -> torch.Tensor:
    if tacs.ndim == 1:
        return tacs.unsqueeze(-1)
    if tacs.ndim != 2:
        raise ValueError(f"tacs must be 1D or 2D. Got shape {tuple(tacs.shape)}.")
    return tacs


def convert_time_to_minutes(
    t: torch.Tensor, *, time_unit: str, warn_on_long_minutes: bool = False
) -> torch.Tensor:
    unit = time_unit.lower()
    if unit in {"min", "minute", "minutes", "m"}:
        if warn_on_long_minutes and t.max().item() > 60.0:
            warnings.warn(
                "time_unit='min' and max(time) > 60. assuming minutes and not "
                "rescaling. If your input is seconds, pass time_unit='s'.",
                stacklevel=2,
            )
        return t
    if unit in {"s", "sec", "second", "seconds"}:
        return t / 60.0
    raise ValueError("time_unit must be either 'min' or 's'.")



import logging
from pathlib import Path

import numpy as np
from rich.progress import Progress

from .._omegaconf_interpreter import sample_config
from ..kinetic.models import make_compartment_model

from .curves import gaussian_noise, generate_aif, generate_time
from .io import SimulationParquetWriter, batch_to_dataframe
from .measured_inputs import MeasuredAIFPool, load_measured_aif_pool


def _get_measured_pool(cfg, *, config_dir: Path | None) -> MeasuredAIFPool | None:
    aif_cfg = cfg.get("aif")
    if aif_cfg is None:
        return None

    measured_files = aif_cfg.get("measured_files")
    if not measured_files:
        return None

    entries = [str(p) for p in measured_files]
    pool = load_measured_aif_pool(entries, base_dir=config_dir)
    logging.info("Loaded %s measured AIF/TIME vectors from config.", pool.n_items)
    return pool


def _infer_synthetic_frame_count(cfg) -> int:
    protocols = cfg.time.protocols
    frame_counts = []
    for _, spec in protocols.items():
        frame_counts.append(sum(int(chunk[0]) for chunk in spec))

    unique_counts = sorted(set(frame_counts))
    if len(unique_counts) != 1:
        raise ValueError(
            "All configured protocols must have the same number of frames. "
            f"Found frame counts: {unique_counts}"
        )
    return unique_counts[0]


def run_simulations(
    cfg,
    *,
    out_dir,
    batch_size=10_000,
    config_dir: Path | None = None,
):
    rng = np.random.default_rng(cfg.seed)
    ignore = ("time.frame_durations_s",)
    total = cfg.num_simulations
    measured_pool = _get_measured_pool(cfg, config_dir=config_dir)
    n_frames = (
        measured_pool.n_frames
        if measured_pool is not None
        else _infer_synthetic_frame_count(cfg)
    )

    with Progress() as progress, SimulationParquetWriter(out_dir=out_dir) as writer:
        task = progress.add_task("[cyan]Simulating...", total=total)
        batch_idx = 0
        sample_id0 = 0
        for start in range(0, total, batch_size):
            n = min(batch_size, total - start)

            batch = {}
            batch["batch_id"] = np.full(n, batch_idx, dtype=np.int32)
            batch["sample_id"] = np.arange(sample_id0, sample_id0 + n, dtype=np.int64)
            batch["protocol"] = np.empty(n, dtype=object)
            batch["aif_source"] = np.empty(n, dtype=object)
            batch["regime"] = np.empty(n, dtype=object)
            batch["noise"] = np.empty(n, dtype=np.float32)
            batch["n_frames_used"] = np.empty(n, dtype=np.int32)
            batch["K1"] = np.empty(n, dtype=np.float32)
            batch["k2"] = np.empty(n, dtype=np.float32)
            batch["k3"] = np.empty(n, dtype=np.float32)
            batch["k4"] = np.empty(n, dtype=np.float32)
            batch["vB"] = np.empty(n, dtype=np.float32)
            batch["Tinj_s"] = np.full(n, np.nan, dtype=np.float32)
            batch["tau"] = np.full(n, np.nan, dtype=np.float32)
            batch["A1"] = np.full(n, np.nan, dtype=np.float32)
            batch["r2"] = np.full(n, np.nan, dtype=np.float32)
            batch["r3"] = np.full(n, np.nan, dtype=np.float32)
            batch["L1"] = np.full(n, np.nan, dtype=np.float32)
            batch["L2"] = np.full(n, np.nan, dtype=np.float32)
            batch["L3"] = np.full(n, np.nan, dtype=np.float32)

            batch["time"] = np.empty((n, n_frames), dtype=np.float32)
            batch["aif"] = np.empty((n, n_frames), dtype=np.float32)
            batch["tac"] = np.empty((n, n_frames), dtype=np.float32)

            for i in range(n):
                sampled = sample_config(cfg, rng, ignore_paths=ignore)

                time_params = sampled["time"]
                scan_duration = time_params["scan_duration"]
                protocol = time_params["protocol"]

                cm_cfg = sampled["compartment_model"]
                regime = cm_cfg["regime"]
                cm_params = cm_cfg["regimes"][regime]
                cm_choice = cm_cfg["num_k"]
                cm_noise = cm_cfg["noise"]

                if measured_pool is None:
                    (
                        t_fine_min,
                        t_fine_s,
                        time_min,
                        frame_durations_s,
                        dt_s,
                    ) = generate_time(
                        frame_durations_s=time_params["protocols"][protocol],
                        dt_s=time_params["dt_s"],
                    )
                    aif_params = sampled["aif"]["params"]
                    aif = generate_aif(
                        t_fine_min,
                        t_fine_s,
                        frame_durations_s,
                        dt_s,
                        **aif_params,
                    )
                    aif_source = "__simulated__"
                else:
                    aif, time_min, aif_source = measured_pool.sample(rng)
                    protocol = "measured"

                compartment_model = make_compartment_model(cm_choice)
                tac = compartment_model(aif, time_min, **cm_params)
                tac = gaussian_noise(tac, rng, factor=cm_noise)

                n_frames_used = np.sum(time_min < scan_duration)

                batch["protocol"][i] = protocol
                batch["aif_source"][i] = aif_source
                batch["regime"][i] = regime
                batch["noise"][i] = cm_noise
                batch["n_frames_used"][i] = n_frames_used
                batch["K1"][i] = cm_params["K1"]
                batch["k2"][i] = cm_params["k2"]
                batch["k3"][i] = cm_params["k3"]
                batch["k4"][i] = cm_params["k4"]
                batch["vB"][i] = cm_params["vB"]
                if measured_pool is None:
                    batch["Tinj_s"][i] = aif_params["Tinj_s"]
                    batch["tau"][i] = aif_params["tau"]
                    batch["A1"][i] = aif_params["A1"]
                    batch["r2"][i] = aif_params["r2"]
                    batch["r3"][i] = aif_params["r3"]
                    batch["L1"][i] = aif_params["L1"]
                    batch["L2"][i] = aif_params["L2"]
                    batch["L3"][i] = aif_params["L3"]

                batch["time"][i] = time_min
                batch["aif"][i] = aif
                batch["tac"][i] = tac

                progress.update(task, advance=1)

            frame = batch_to_dataframe(batch)
            writer.write_batch(frame)
            del batch

            n_batches = (total + batch_size - 1) // batch_size
            logging.debug("%s/%s batches complete...", batch_idx + 1, n_batches)

            sample_id0 += n
            batch_idx += 1



"""Compatibility facade for kinetic modeling.

Implementation now lives in the `lls.kinetic` subpackage.
"""

from .kinetic.high_level import fit_compartment_model
from .kinetic.io_utils import (
    convert_time_to_minutes as _convert_time_to_minutes,
    infer_device as _infer_device,
    infer_dtype as _infer_dtype,
    normalize_tacs_shape as _normalize_tacs_shape,
    require_1d as _require_1d,
    to_tensor as _to_tensor,
)
from .kinetic.math_utils import (
    build_interpolation_grid as _build_interpolation_grid,
    interp1d_linear as _interp1d_linear,
    quadratic_formula,
    trapezoidal_integral,
)
from .kinetic.models import (
    CompartmentModel,
    ForwardCompartmentModel,
    Irreversible3KModel,
    Irreversible3KModelvB,
    OneTissue2KModel,
    Reversible4KModel,
    Reversible4KModelvB,
    make_compartment_model,
)
from .kinetic.solver import lls_3k, lls_3k_vB, lls_4k, lls_4k_vB
from .kinetic.types import ArrayLike, DeviceLike, Scalar, VBLike

__all__ = [
    "lls_3k",
    "lls_3k_vB",
    "lls_4k",
    "lls_4k_vB",
    "fit_compartment_model",
    "ForwardCompartmentModel",
    "CompartmentModel",
    "OneTissue2KModel",
    "Irreversible3KModel",
    "Irreversible3KModelvB",
    "Reversible4KModel",
    "Reversible4KModelvB",
    "make_compartment_model",
    "trapezoidal_integral",
    "quadratic_formula",
    "Scalar",
    "VBLike",
    "DeviceLike",
    "ArrayLike",
]



from importlib import import_module

from .kinetic_modeling import (
    fit_compartment_model,
    lls_3k,
    lls_3k_vB,
    lls_4k,
    lls_4k_vB,
    make_compartment_model,
)

__all__ = [
    "make_compartment_model",
    "lls_3k",
    "lls_3k_vB",
    "lls_4k",
    "lls_4k_vB",
    "fit_compartment_model",
]

_OPTIONAL_EXPORTS = {
    "sample_config": ("._omegaconf_interpreter", "sample_config"),
    "make_table": ("._parquet_shards", "make_table"),
    "write_parquet_dataset": ("._parquet_shards", "write_parquet_dataset"),
    "write_time_sidecar": ("._parquet_shards", "write_time_sidecar"),
    "batch_to_table": ("._parquet_shards", "batch_to_table"),
    "write_batch": ("._parquet_shards", "write_batch"),
}

_OPTIONAL_DEPENDENCIES = {"omegaconf", "pandas", "pyarrow", "rich"}


def __getattr__(name: str):
    target = _OPTIONAL_EXPORTS.get(name)
    if target is None:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    module_name, attr_name = target
    try:
        module = import_module(module_name, __name__)
    except ModuleNotFoundError as exc:
        missing = exc.name.split(".")[0] if exc.name else None
        if missing is None:
            message = str(exc)
            for dependency in _OPTIONAL_DEPENDENCIES:
                if dependency in message:
                    missing = dependency
                    break

        if missing in _OPTIONAL_DEPENDENCIES:
            raise ModuleNotFoundError(
                f"`lls.{name}` requires optional simulation/analysis dependencies. "
                "Install them with `lls[sim]`."
            ) from exc
        raise

    value = getattr(module, attr_name)
    globals()[name] = value
    return value



"""Backward-compatible forward-model facade.

Forward model implementation now lives in ``lls.kinetic.models`` so both
forward and inverse model code are colocated.
"""

from .kinetic.models import make_compartment_model

__all__ = ["make_compartment_model"]



from __future__ import annotations

import glob
from dataclasses import dataclass
from pathlib import Path

import numpy as np


@dataclass(frozen=True)
class MeasuredAIFPool:
    aif: np.ndarray
    time: np.ndarray
    sources: tuple[str, ...]

    @property
    def n_items(self) -> int:
        return int(self.aif.shape[0])

    @property
    def n_frames(self) -> int:
        return int(self.aif.shape[1])

    def sample(self, rng: np.random.Generator) -> tuple[np.ndarray, np.ndarray, str]:
        idx = int(rng.integers(0, self.n_items))
        return self.aif[idx], self.time[idx], self.sources[idx]


def resolve_measured_files(
    entries: list[str] | tuple[str, ...], *, base_dir: Path | None
) -> list[Path]:
    paths: list[Path] = []
    for entry in entries:
        raw = str(entry)
        path = Path(raw)
        if not path.is_absolute() and base_dir is not None:
            path = base_dir / path

        if any(ch in str(path) for ch in "*?[]"):
            matched = [Path(p) for p in sorted(glob.glob(str(path)))]
            paths.extend(matched)
            continue

        if path.is_dir():
            paths.extend(sorted(path.glob("*.npz")))
            continue

        paths.append(path)

    return paths


def _load_single_aif_file(path: Path) -> tuple[np.ndarray, np.ndarray]:
    if not path.exists():
        raise FileNotFoundError(f"Measured AIF file not found: {path}")

    with np.load(path) as data:
        if "AIF" not in data or "TIME" not in data:
            raise KeyError(
                f"{path} must contain 'AIF' and 'TIME' arrays (npz keys)."
            )
        aif = np.asarray(data["AIF"], dtype=np.float32).reshape(-1)
        time = np.asarray(data["TIME"], dtype=np.float32).reshape(-1)

    if aif.size < 2:
        raise ValueError(f"{path} has too few frames. Expected at least 2.")
    if aif.shape != time.shape:
        raise ValueError(
            f"{path} has mismatched shapes: AIF={aif.shape}, TIME={time.shape}."
        )
    if not np.isfinite(aif).all() or not np.isfinite(time).all():
        raise ValueError(f"{path} contains non-finite values.")
    if np.any(time[1:] <= time[:-1]):
        raise ValueError(f"{path} TIME must be strictly increasing.")

    return aif, time


def load_measured_aif_pool(
    entries: list[str] | tuple[str, ...], *, base_dir: Path | None = None
) -> MeasuredAIFPool:
    files = resolve_measured_files(entries, base_dir=base_dir)
    if not files:
        raise ValueError(
            "No measured AIF files resolved from config 'aif.measured_files'."
        )

    aif_list: list[np.ndarray] = []
    time_list: list[np.ndarray] = []
    sources: list[str] = []

    expected_frames: int | None = None
    for path in files:
        aif, time = _load_single_aif_file(path)
        n_frames = int(aif.shape[0])
        if expected_frames is None:
            expected_frames = n_frames
        elif n_frames != expected_frames:
            raise ValueError(
                "All measured AIF/TIME vectors must have the same frame count. "
                f"Expected {expected_frames}, got {n_frames} for {path}."
            )

        aif_list.append(aif)
        time_list.append(time)
        sources.append(str(path))

    aif_arr = np.stack(aif_list, axis=0).astype(np.float32, copy=False)
    time_arr = np.stack(time_list, axis=0).astype(np.float32, copy=False)
    return MeasuredAIFPool(aif=aif_arr, time=time_arr, sources=tuple(sources))




from pathlib import Path

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

CURVE_KEYS = ("aif", "tac", "time")


def batch_to_dataframe(batch: dict[str, np.ndarray]) -> pd.DataFrame:
    n = batch["tac"].shape[0]
    cols = {}

    for key, value in batch.items():
        if key in CURVE_KEYS:
            continue
        arr = np.asarray(value)
        if arr.shape[0] != n:
            raise ValueError(
                f"Column '{key}' has {arr.shape[0]} rows, expected {n} rows."
            )
        cols[key] = arr

    for key in CURVE_KEYS:
        curve = np.asarray(batch[key], dtype=np.float32, order="C")
        if curve.ndim != 2 or curve.shape[0] != n:
            raise ValueError(
                f"Curve '{key}' must have shape (N, T). Got {tuple(curve.shape)}."
            )
        t = curve.shape[1]
        for j in range(t):
            cols[f"{key}_{j:03d}"] = curve[:, j]

    return pd.DataFrame(cols)


class SimulationParquetWriter:
    def __init__(
        self,
        out_dir,
        *,
        file_name: str = "simulations.parquet",
        compression: str = "zstd",
    ):
        self.out_dir = Path(out_dir)
        self.path = self.out_dir / file_name
        self.compression = compression
        self._writer = None

    def __enter__(self):
        self.out_dir.mkdir(parents=True, exist_ok=True)
        if self.path.exists():
            self.path.unlink()
        return self

    def write_batch(self, frame: pd.DataFrame):
        table = pa.Table.from_pandas(frame, preserve_index=False)
        if self._writer is None:
            self._writer = pq.ParquetWriter(
                str(self.path),
                table.schema,
                compression=self.compression,
                use_dictionary=True,
                write_statistics=True,
            )
        self._writer.write_table(table)

    def close(self):
        if self._writer is not None:
            self._writer.close()
            self._writer = None

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


__all__ = ["SimulationParquetWriter", "batch_to_dataframe"]



from abc import ABC, abstractmethod
from typing import ClassVar

import numpy as np
import torch

from .math_utils import quadratic_formula, trapezoidal_integral
from .types import DeviceLike


class ForwardCompartmentModel(ABC):
    """Forward model contract used for TAC simulation."""

    @abstractmethod
    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray: ...

    def forward_cpet(
        self,
        Cp: np.ndarray,
        t: np.ndarray,
        K1: float,
        k2: float,
        vB: float,
        k3: float = 0.0,
        k4: float = 0.0,
    ) -> np.ndarray:
        cp = np.asarray(Cp)
        time = np.asarray(t)
        h = self.impulse_response(time, K1=K1, k2=k2, k3=k3, k4=k4)
        ct = np.convolve(h, cp, mode="full") * (time[1] - time[0])
        ct = ct[: len(time)]
        return (1 - vB) * ct + vB * cp


class CompartmentModel(ForwardCompartmentModel, ABC):
    """Model-specific hooks for design matrix construction and parameter recovery."""

    n_linear_params: ClassVar[int] = 0
    fits_vB: ClassVar[bool] = False

    @abstractmethod
    def build_design_matrix(
        self, Ct: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor
    ) -> torch.Tensor: ...

    @abstractmethod
    def coefficient_bounds(
        self, *, device: DeviceLike | None, dtype: torch.dtype | str | None
    ) -> tuple[torch.Tensor, torch.Tensor]: ...

    @abstractmethod
    def recover_kinetic_parameters(
        self, p: torch.Tensor, eps: float
    ) -> torch.Tensor: ...


class Irreversible3KModel(CompartmentModel):
    n_linear_params: ClassVar[int] = 3

    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray:
        del k4
        return _k3_impulse_response(t, K1=K1, k2=k2, k3=k3)

    def build_design_matrix(
        self, Ct: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor
    ) -> torch.Tensor:
        T, N = Ct.shape

        # Compute integrals for the design matrix
        I_Ct = trapezoidal_integral(Ct, t)  # (T, N)
        I_Cp = trapezoidal_integral(Cp, t)  # (T,)
        I2_Cp = trapezoidal_integral(I_Cp, t)  # (T,)

        # Design matrix: (T, N, 3) with columns [I_Ct, I_Cp, I2_Cp]
        X = torch.empty((T, N, 3), dtype=Ct.dtype, device=Ct.device)
        X[..., 0] = I_Ct
        X[..., 1] = I_Cp.unsqueeze(-1)
        X[..., 2] = I2_Cp.unsqueeze(-1)
        return X

    def coefficient_bounds(
        self, *, device: DeviceLike | None, dtype: torch.dtype | str | None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Bounds assume mapping:
        #   p1 <= 0,  p2 >= 0,  p3 >= 0
        p_lower = torch.tensor([-torch.inf, 0.0, 0.0], device=device, dtype=dtype)
        p_upper = torch.tensor([0.0, torch.inf, torch.inf], device=device, dtype=dtype)
        return p_lower, p_upper

    def recover_kinetic_parameters(self, p: torch.Tensor, eps: float) -> torch.Tensor:
        p1, p2, p3 = p[:, 0], p[:, 1], p[:, 2]

        # Recover kinetic parameters from linear coefficients
        K1 = p2
        k3 = p3 / (K1 + eps)
        k2 = -k3 - p1
        Ki = K1 * k3 / (k2 + k3 + eps)

        return torch.stack([K1, k2, k3, Ki], dim=0)


class Irreversible3KModelvB(CompartmentModel):
    n_linear_params: ClassVar[int] = 4
    fits_vB: ClassVar[bool] = True

    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray:
        del k4
        return _k3_impulse_response(t, K1=K1, k2=k2, k3=k3)

    def build_design_matrix(
        self, Ct: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor
    ) -> torch.Tensor:
        T, N = Ct.shape

        # Compute integrals for the design matrix
        I_Ct = trapezoidal_integral(Ct, t)  # (T, N)
        I_Cp = trapezoidal_integral(Cp, t)  # (T,)
        I2_Cp = trapezoidal_integral(I_Cp, t)  # (T,)

        # Design matrix: (T, N, 4) with columns [I_Ct, Cp, I_Cp, I2_Cp]
        X = torch.empty((T, N, 4), dtype=Ct.dtype, device=Ct.device)
        X[..., 0] = I_Ct
        X[..., 1] = Cp.unsqueeze(-1)
        X[..., 2] = I_Cp.unsqueeze(-1)
        X[..., 3] = I2_Cp.unsqueeze(-1)
        return X

    def coefficient_bounds(
        self, *, device: DeviceLike | None, dtype: torch.dtype | str | None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Bounds assume mapping:
        #   p1 <= 0, 0 <= p2 <= 1,  p3 >= 0,  p4 >= 0
        p_lower = torch.tensor([-torch.inf, 0.0, 0.0, 0.0], device=device, dtype=dtype)
        p_upper = torch.tensor(
            [0.0, 1.0, torch.inf, torch.inf], device=device, dtype=dtype
        )
        return p_lower, p_upper

    def recover_kinetic_parameters(self, p: torch.Tensor, eps: float) -> torch.Tensor:
        p1, p2, p3, p4 = p[:, 0], p[:, 1], p[:, 2], p[:, 3]

        # Recover kinetic parameters from linear coefficients
        K1 = (p3 + p1 * p2) / (1 - p2 + eps)
        k3 = p4 / ((1 - p2) * K1 + eps)
        k2 = -k3 - p1
        vB = p2
        Ki = K1 * k3 / (k2 + k3 + eps)

        return torch.stack([K1, k2, k3, vB, Ki], dim=0)


class Reversible4KModel(CompartmentModel):
    n_linear_params: ClassVar[int] = 4

    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray:
        return _k4_impulse_response(t, K1=K1, k2=k2, k3=k3, k4=k4)

    def build_design_matrix(
        self, Ct: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor
    ) -> torch.Tensor:
        T, N = Ct.shape

        # Compute integrals for the design matrix
        I_Ct = trapezoidal_integral(Ct, t)  # (T, N)
        I2_Ct = trapezoidal_integral(I_Ct, t)  # (T, N)
        I_Cp = trapezoidal_integral(Cp, t)  # (T,)
        I2_Cp = trapezoidal_integral(I_Cp, t)  # (T,)

        # Design matrix: (T, N, 4) with columns [I_Ct, I2_Ct, I_Cp, I2_Cp]
        X = torch.empty((T, N, 4), dtype=Ct.dtype, device=Ct.device)
        X[..., 0] = I_Ct
        X[..., 1] = I2_Ct
        X[..., 2] = I_Cp.unsqueeze(-1)
        X[..., 3] = I2_Cp.unsqueeze(-1)
        return X

    def coefficient_bounds(
        self, *, device: DeviceLike | None, dtype: torch.dtype | str | None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Bounds assume mapping:
        #   p1 <= 0,  p2 <= 0,  p3 >= 0, p4 >= 0
        p_lower = torch.tensor(
            [-torch.inf, -torch.inf, 0.0, 0.0], device=device, dtype=dtype
        )
        p_upper = torch.tensor(
            [0.0, 0.0, torch.inf, torch.inf], device=device, dtype=dtype
        )
        return p_lower, p_upper

    def recover_kinetic_parameters(self, p: torch.Tensor, eps: float) -> torch.Tensor:
        p1, p2, p3, p4 = p[:, 0], p[:, 1], p[:, 2], p[:, 3]

        # Recover kinetic parameters from linear coefficients
        K1 = p3
        s = p4 / (p3 + eps)
        k2 = -p1 - s
        k4 = -p2 / (k2 + eps)
        k3 = s - k4
        Ki = K1 * k3 / (k2 + k3 + eps)

        return torch.stack([K1, k2, k3, k4, Ki], dim=0)


class Reversible4KModelvB(CompartmentModel):
    n_linear_params: ClassVar[int] = 5
    fits_vB: ClassVar[bool] = True

    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray:
        return _k4_impulse_response(t, K1=K1, k2=k2, k3=k3, k4=k4)

    def build_design_matrix(
        self, Ct: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor
    ) -> torch.Tensor:
        T, N = Ct.shape

        # Compute integrals for the design matrix
        I_Ct = trapezoidal_integral(Ct, t)  # (T, N)
        I2_Ct = trapezoidal_integral(I_Ct, t)  # (T, N)
        I_Cp = trapezoidal_integral(Cp, t)  # (T,)
        I2_Cp = trapezoidal_integral(I_Cp, t)  # (T,)

        # Design matrix: (T, N, 5) with columns [I_Ct, I2_Ct, Cp, I_Cp, I2_Cp]
        X = torch.empty((T, N, 5), dtype=Ct.dtype, device=Ct.device)
        X[..., 0] = I_Ct
        X[..., 1] = I2_Ct
        X[..., 2] = Cp.unsqueeze(-1)
        X[..., 3] = I_Cp.unsqueeze(-1)
        X[..., 4] = I2_Cp.unsqueeze(-1)
        return X

    def coefficient_bounds(
        self, *, device: DeviceLike | None, dtype: torch.dtype | str | None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Bounds assume mapping:
        #   p1 <= 0,  p2 <= 0, 0 <= p3 <= 1, p4 >= 0, p5 >= 0
        p_lower = torch.tensor(
            [-torch.inf, -torch.inf, 0.0, 0.0, 0.0], device=device, dtype=dtype
        )
        p_upper = torch.tensor(
            [0.0, 0.0, 1.0, torch.inf, torch.inf], device=device, dtype=dtype
        )
        return p_lower, p_upper

    def recover_kinetic_parameters(self, p: torch.Tensor, eps: float) -> torch.Tensor:
        p1, p2, p3, p4, p5 = p[:, 0], p[:, 1], p[:, 2], p[:, 3], p[:, 4]

        # Recover kinetic parameters from linear coefficients
        K1 = (p4 + p1 * p3) / (1 - p3 + eps)
        k3 = p5 / (p4 + p1 * p3 + eps) + p3 * p2

        s1, s2 = quadratic_formula(1, p1 + k3, -p2)
        k2_mask = s1 > s2
        k2 = torch.where(k2_mask, s1, s2)
        k4 = torch.where(k2_mask, s2, s1)
        vB = p3

        Ki = K1 * k3 / (k2 + k3 + eps)

        return torch.stack([K1, k2, k3, k4, vB, Ki], dim=0)


class OneTissue2KModel(ForwardCompartmentModel):
    def impulse_response(
        self, t: np.ndarray, K1: float, k2: float, k3: float = 0.0, k4: float = 0.0
    ) -> np.ndarray:
        del k3, k4
        return _k2_impulse_response(t, K1=K1, k2=k2)


MODEL_2K = OneTissue2KModel()
MODEL_3K = Irreversible3KModel()
MODEL_3K_vB = Irreversible3KModelvB()
MODEL_4K = Reversible4KModel()
MODEL_4K_vB = Reversible4KModelvB()


def _k2_impulse_response(t: np.ndarray, K1: float, k2: float) -> np.ndarray:
    return K1 * np.exp(-k2 * t)


def _k3_impulse_response(
    t: np.ndarray, K1: float, k2: float, k3: float
) -> np.ndarray:
    return K1 / (k2 + k3) * (k3 + k2 * np.exp(-(k2 + k3) * t))


def _k4_impulse_response(
    t: np.ndarray, K1: float, k2: float, k3: float, k4: float
) -> np.ndarray:
    disc = np.sqrt((k2 + k3 + k4) ** 2 - 4 * k2 * k4)
    a1 = 0.5 * (k2 + k3 + k4 + disc)
    a2 = 0.5 * (k2 + k3 + k4 - disc)

    b1 = (a1 - k3 - k4) * np.exp(-(a1 * t))
    b2 = (a2 - k3 - k4) * np.exp(-(a2 * t))

    return K1 / (a1 - a2) * (b1 - b2)


def make_compartment_model(num_k: int = 2):
    """Return forward compartment-model callable used for TAC simulation.

    The returned function maps (Cp, t, kinetic params) -> Cpet.
    """
    forward_models = {
        2: MODEL_2K,
        3: MODEL_3K,
        4: MODEL_4K,
    }
    model = forward_models.get(num_k)
    if model is None:
        def unsupported(*args, **kwargs):
            raise NotImplementedError(
                "Only compartment models with num_k in {2, 3, 4} are supported."
            )
        return unsupported
    return model.forward_cpet



import warnings

import torch

from .models import (
    MODEL_3K,
    MODEL_4K,
    CompartmentModel,
    MODEL_3K_vB,
    MODEL_4K_vB,
)
from .types import DeviceLike, VBLike


def _validate_inputs(Cpet: torch.Tensor, Cp: torch.Tensor, t: torch.Tensor) -> None:
    assert Cpet.device == Cp.device == t.device, "All inputs must be on the same device"
    assert Cpet.dtype == Cp.dtype == t.dtype, "All inputs must have the same dtype"
    assert Cpet.ndim == 2, "Cpet must have shape (T, N)"
    assert Cp.ndim == 1, "Cp must have shape (T,)"
    assert t.ndim == 1, "t must have shape (T,)"
    assert Cpet.size(0) == Cp.size(0) == t.size(0), (
        "Mismatched time dimension in inputs"
    )


def _prepare_tissue_concentration(
    Cpet: torch.Tensor, Cp: torch.Tensor, vB: VBLike, eps: float
) -> tuple[torch.Tensor, torch.Tensor]:
    # Convert Cpet to Ct (tissue concentration) by removing blood contribution
    device = Cpet.device
    dtype = Cpet.dtype
    n_voxels = Cpet.size(1)

    vB_t = vB if torch.is_tensor(vB) else torch.tensor(vB, device=device, dtype=dtype)
    if vB_t.ndim == 0:  # Scalar
        vB_t = vB_t.clamp(0.0, 1.0 - eps)
        den = (1.0 - vB_t).clamp_min(eps)
        Ct = (Cpet - vB_t * Cp[:, None]) / den
    else:  # voxelwise (N,)
        vB_t = vB_t.to(device=device, dtype=dtype).clamp(0.0, 1.0 - eps)
        assert vB_t.ndim == 1 and vB_t.numel() == n_voxels, (
            "Vector vB must have shape (N,)"
        )
        den = (1.0 - vB_t).clamp_min(eps)  # (N,)
        Ct = (Cpet - Cp[:, None] * vB_t[None, :]) / den[None, :]

    return Ct, vB_t


def _solve_unconstrained(
    Gram: torch.Tensor, Y: torch.Tensor, *, device: DeviceLike
) -> torch.Tensor:
    try:
        return torch.linalg.solve(Gram, Y)
    except Exception as e:
        warnings.warn(f"Solve failed on {device} ({e}). Falling back to CPU float 64.")
        Gram_cpu = Gram.double().cpu()
        Y_cpu = Y.double().cpu()
        p_cpu = torch.linalg.solve(Gram_cpu, Y_cpu).float()
        return p_cpu.to(device)


def _solve_coefficients(
    X: torch.Tensor,
    Ct: torch.Tensor,
    *,
    reg: float,
    eps: float,
    column_scale: bool,
    nnls: bool,
    nnls_max_iter: int,
    nnls_tol: float,
    p_lower: torch.Tensor,
    p_upper: torch.Tensor,
) -> torch.Tensor:
    _, N, P = X.shape
    device = X.device
    dtype = X.dtype

    # Column scaling
    if column_scale:
        # per-voxel scaling for all columns (simple + robust)
        col_scale = torch.sqrt(torch.sum(X * X, dim=0)).clamp_min(eps)  # (N, P)
        Xs = X / col_scale[None, :, :]
    else:
        col_scale = None
        Xs = X

    # Gram matrix: X^T @ X
    Gram = torch.einsum("tni,tnj->nij", Xs, Xs)  # (N, P, P)

    # Apply Tikhonov regularization (ridge) to the Gram matrix
    Gram += reg * torch.eye(P, dtype=dtype, device=device).expand(N, P, P)

    # Right-hand side: X^T @ Ct
    Y = torch.einsum("tni,tn->ni", Xs, Ct)  # (N, P)

    # Non negative least squares with projected gradient descent
    if nnls:
        assert nnls_max_iter >= 1, "nnls_max_iter must be >= 1"
        assert nnls_tol > 0, "nnls_tol must be > 0"

        lower = p_lower.view(1, P).expand(N, P)
        upper = p_upper.view(1, P).expand(N, P)

        def project(p: torch.Tensor) -> torch.Tensor:
            return torch.minimum(torch.maximum(p, lower), upper)

        p = project(_solve_unconstrained(Gram, Y, device=device))

        # Step size from ||Gram||_inf upper bound of Lipschitz constant
        L_inf = (
            torch.abs(Gram).sum(dim=2).amax(dim=1, keepdim=True).clamp_min(eps)
        )  # (N,1)
        step = 1.0 / L_inf

        # Projected gradient iterations
        p_norm_floor = 1.0
        for _ in range(nnls_max_iter):
            grad = torch.bmm(Gram, p.unsqueeze(-1)).squeeze(-1) - Y  # (N, P)
            p_next = project(p - step * grad)

            # convergence
            rel = (p_next - p).norm(dim=1) / p.norm(dim=1).clamp_min(p_norm_floor)
            p = p_next
            if torch.max(rel) <= nnls_tol:
                break

        if not torch.isfinite(p).all():
            warnings.warn(
                "NNLS produced non-finite p values; consider increasing reg or enabling column_scale."
            )
    else:
        p = _solve_unconstrained(Gram, Y, device=device)

    # Unscale coefficients back if we scaled columns
    if column_scale:
        p = p / col_scale

    return p


def _lls_solve(
    Cpet: torch.Tensor,
    Cp: torch.Tensor,
    t: torch.Tensor,
    *,
    vB: VBLike | None,
    model: CompartmentModel,
    reg: float,
    eps: float,
    column_scale: bool,
    nnls: bool,
    nnls_max_iter: int,
    nnls_tol: float,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    _validate_inputs(Cpet, Cp, t)

    # Fixed-vB models solve in Ct-space, while fitted-vB models solve in Cpet-space.
    if model.fits_vB:
        y = Cpet
        X = model.build_design_matrix(Cpet, Cp, t)
        vB_t = None
    else:
        y, vB_t = _prepare_tissue_concentration(Cpet, Cp, vB, eps)
        X = model.build_design_matrix(y, Cp, t)

    p_lower, p_upper = model.coefficient_bounds(device=Cpet.device, dtype=Cpet.dtype)

    p = _solve_coefficients(
        X,
        y,
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
        p_lower=p_lower,
        p_upper=p_upper,
    )

    if model.fits_vB:
        # Compute fitted Cpet directly in Cpet-space
        fitted_Cpet = torch.einsum("tni,ni->tn", X, p)  # (T, N)
    else:
        # Compute fitted tissue concentration at each time point: X @ p
        fitted_Ct = torch.einsum("tni,ni->tn", X, p)  # (T, N)

        # Add back blood contribution to get fitted Cpet
        fitted_Cpet = fitted_Ct * (1.0 - vB_t) + Cp[:, None] * vB_t

    mse = torch.mean(torch.square(fitted_Cpet - Cpet), dim=0)  # (N,)

    kinetic_params = model.recover_kinetic_parameters(p, eps)
    return kinetic_params, fitted_Cpet, mse


def lls_3k(
    Cpet: torch.Tensor,
    Cp: torch.Tensor,
    t: torch.Tensor,
    *,
    vB: VBLike,
    reg: float = 1e-6,
    eps: float = 1e-8,
    column_scale: bool = True,
    nnls: bool = False,
    nnls_max_iter: int = 200,
    nnls_tol: float = 1e-6,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Linear least squares fitting of the irreversible two-tissue compartment model."""
    return _lls_solve(
        Cpet,
        Cp,
        t,
        vB=vB,
        model=MODEL_3K,
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
    )


def lls_4k(
    Cpet: torch.Tensor,
    Cp: torch.Tensor,
    t: torch.Tensor,
    *,
    vB: VBLike,
    reg: float = 1e-6,
    eps: float = 1e-8,
    column_scale: bool = True,
    nnls: bool = False,
    nnls_max_iter: int = 200,
    nnls_tol: float = 1e-6,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Linear least squares fitting of the reversible two-tissue compartment model."""
    return _lls_solve(
        Cpet,
        Cp,
        t,
        vB=vB,
        model=MODEL_4K,
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
    )


def lls_3k_vB(
    Cpet: torch.Tensor,
    Cp: torch.Tensor,
    t: torch.Tensor,
    *,
    reg: float = 1e-6,
    eps: float = 1e-8,
    column_scale: bool = True,
    nnls: bool = False,
    nnls_max_iter: int = 200,
    nnls_tol: float = 1e-6,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Linear least squares fitting of the irreversible model with fitted blood fraction."""
    return _lls_solve(
        Cpet,
        Cp,
        t,
        vB=None,
        model=MODEL_3K_vB,
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
    )


def lls_4k_vB(
    Cpet: torch.Tensor,
    Cp: torch.Tensor,
    t: torch.Tensor,
    *,
    reg: float = 1e-6,
    eps: float = 1e-8,
    column_scale: bool = True,
    nnls: bool = False,
    nnls_max_iter: int = 200,
    nnls_tol: float = 1e-6,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Linear least squares fitting of the reversible model with fitted blood fraction."""
    return _lls_solve(
        Cpet,
        Cp,
        t,
        vB=None,
        model=MODEL_4K_vB,
        reg=reg,
        eps=eps,
        column_scale=column_scale,
        nnls=nnls,
        nnls_max_iter=nnls_max_iter,
        nnls_tol=nnls_tol,
    )
